# Provides an example of a self-navigating algorithm which you are under no obligation to use

from microprediction.writer import MicroWriter
from microprediction.reader import default_url
import muid, random, time
from microprediction.samplers import exponential_bootstrap
import pprint

class MicroCrawler(MicroWriter):

    def __init__(self, write_key=None, base_url=None, verbose=True, min_lagged=50, min_difficulty=12, sleep_time=300) :
        """ Initialize write_key if none provided """
        super().__init__(base_url=base_url or default_url(), write_key=write_key )
        assert muid.difficulty(write_key) >= 8, "Invalid write_key for crawler. See www.muid.org to mine one. "
        self.verbose = verbose
        self.min_lagged = min_lagged          # Only consider streams with a long lag history
        self.min_difficulty = min_difficulty  # Only choose streams with sponsors at least this long
        self.sleep_time = sleep_time          # Seconds to sleep between actions

    def choose_stream(self):
        """ Randomized """
        streams = self.get_streams()
        candidates = [ name for name, sponsor in streams.items() if len(sponsor.replace(' ',''))>=self.min_difficulty ]
        return random.choice(candidates)

    def choose_horizon(self):
        return random.choice(self.delays)

    def run(self):
        while True:
            name  = self.choose_stream()
            delay = self.choose_horizon()
            lagged = self.get_lagged_values(name)
            if len(lagged)<self.min_lagged:
                message = {'name':name,'submitted':False,"reason":"Insufficient lags","lagged_len":len(lagged)}
            else:
                scenario_values = exponential_bootstrap(lagged=lagged,num=self.num_predictions, decay=0.01)
                self.submit(name=name,values=scenario_values,delay=delay)
                balance = self.get_balance()
                errors  = self.get_errors()
                confirms = self.get_confirms()
                message = {'name': name, "submitted": True, 'delay': delay, "values": scenario_values[:5],
                           "balance": balance, "confirms": confirms, "errors": errors[:5], }
            if self.verbose:
                pprint.pprint(message)
                print("",flush=True)
            time.sleep(self.sleep_time)









# Examples of scenario generation models
import random, bisect
import numpy as np

# --------------------------------------------------------------------------
#          Recency bootstrappy things
# --------------------------------------------------------------------------
# Easy to understand benchmarks. Feel free to contribute more.


def exponential_bootstrap( lagged, decay, num, as_process=None):
    as_process = as_process or is_process(lagged)
    return differenced_bootstrap(lagged=lagged, decay=decay, num=num) if as_process else independent_bootstrap(lagged=lagged, decay=decay, num=num)

def independent_bootstrap(lagged, decay, num):
    """ One parameter jiggled bootstrap favouring more recent observations
          lagged  [ float ]     List most recent observation first
          decay    float        Coefficient in exp(-a k) that weights samples
          num      int          Number of scenarios requested
           :returns  [ float ]  Statistical sample
    """
    weights          = list( np.exp( [ -decay*k for k in range(len(lagged)) ] ) )
    empirical_sample = weighted_random_sample(population=lagged, weights=weights, num=num )
    noise            = np.random.randn(num)
    return [ x + decay*eps for x,eps in zip(empirical_sample, noise) ]

def differenced_bootstrap(lagged, decay, num):
    """ One parameter jiggled bootstrap favouring more recent observations (applied to differences processes) """
    safe_diff_lagged = np.diff( list(lagged)+ [0.,0.])
    diff_samples     = independent_bootstrap( lagged=safe_diff_lagged , decay=decay, num=num )
    return [ lagged[0] + dx for dx in diff_samples ]


# --------------------------------------------------------------------------
#            Diagnostics
# --------------------------------------------------------------------------

def sign_changes( lagged ):
    return np.nansum( [ abs(d)>1.5 for d in np.diff( np.sign( list(lagged) + [0., 0.] )) ] )

def is_process( lagged ):
    return sign_changes( np.diff(lagged) ) > 2*sign_changes( lagged )

# --------------------------------------------------------------------------
#            Gaussian
# --------------------------------------------------------------------------

def gaussian_samples(lagged,num,as_process=None):
    as_process = as_process or is_process(lagged)
    return diff_gaussian_samples(lagged=lagged, num=num) if as_process else independent_gaussian_samples(lagged=lagged, num=num)

def evenly_spaced_percentles(num):
    return [ 1./(2*num) ] + list( 1./(2*num)+np.cumsum( (1/num)*np.ones(num-1) ) )

def independent_gaussian_samples(lagged,num):
    shrunk_std  = np.nanstd( list(lagged)+[0.01, -0.01])
    shrunk_mean = np.nanmean( lagged + [0.0] )
    norminv     = _norminv_function()
    return [ shrunk_mean+shrunk_std*norminv(p) for p in evenly_spaced_percentles(num) ]

def diff_gaussian_samples(lagged, num):
    """ Samples from differences """
    safe_diff_lagged = np.diff( list(lagged)+ [0.,0.])
    diff_samples     = independent_gaussian_samples( lagged=safe_diff_lagged , num=num )
    return [ lagged[0] + dx for dx in diff_samples ]


def _norminv_function():
    try:
        from statistics import NormalDist
        return NormalDist(mu=0, sigma=1.0).inv_cdf
    except ImportError:
        from scipy.stats import norm
        return norm.ppf


# --------------------------------------------------------------------------
#            Helpers
# --------------------------------------------------------------------------


def weighted_random_sample(weights, num, population=None):
    """ Weighted version of random.sample()  """
    wrg  = WeightedRandomGenerator(weights)
    ndx = [ wrg() for _ in range(num) ]
    return ndx if population is None else [ population[k] for k in ndx ]

class WeightedRandomGenerator(object):
    def __init__(self, weights):
        self.totals = []
        running_total = 0

        for w in weights:
            running_total += w
            self.totals.append(running_total)

    def next(self):
        rnd = random.random() * self.totals[-1]
        return bisect.bisect_right(self.totals, rnd)

    def __call__(self):
        return self.next()





import requests

# Number of bikes available at Madison and 48th


def fetch_live_data(key,field):
    r = requests.get("https://feeds.citibikenyc.com/stations/stations.json")
    if r.status_code==200:
        data = r.json()
        station_data = [ d for d in data["stationBeanList"] if int(d["id"])==key ]
        if len(station_data)==1:
            record = station_data[0]
            return record[field]
        else:
            raise Exception("Invalid key")
    else:
        return None


def madison_bikes_available():
    return fetch_live_data(key=456,field="availableBikes")

